#include <iostream>
#include <vector>
#include <stdexcept>
#include <cmath>

class Alakzat {
    protected:
        double x, y;
        unsigned szin;
    public:
        Alakzat(unsigned szin, double x, double y)
            : x(x), y(y), szin(szin) {}
        virtual void kiir() const = 0;
        virtual double kerulet() const = 0;
        virtual Alakzat *klon() const = 0;
        virtual ~Alakzat() {}
};

class Teglalap : public Alakzat {
    private:
        double a, b;
    public:
        Teglalap(unsigned szin, double x, double y, double a, double b)
            : Alakzat(szin, x, y), a(a), b(b) {}
        
        virtual void kiir() const {
            std::cout << "Téglalap x=" << x << " y=" << y
                      << " a=" << a << " b=" << b << std::endl;
        }
        
        virtual double kerulet() const {
            return 2*(a+b);
        }
        
        virtual Alakzat *klon() const {
            return new Teglalap(*this);
        }
};

class Kor : public Alakzat {
    private:
        double r;
    public:
        Kor(unsigned szin, double x, double y, double r)
            : Alakzat(szin, x, y), r(r) {}
        
        virtual void kiir() const {
            std::cout << "Kör x=" << x << " y=" << y
                      << " r=" << r << std::endl;
        }
        
        virtual double kerulet() const {
            return 2*3.1415926535 * r;
        }

        virtual Alakzat *klon() const {
            return new Kor(*this);
        }
};

/* ez nem örököltethető a téglalapból! a téglalap képes széltében
 * és hosszában eltérő méreteket fölvenni, a négyzet nem. tehát
 * a négyzet nem lehet téglalap. a madarak képesek repülni, a strucc
 * nem. tehát a strucc nem madár. biológia órán igen, de oop-ből nem.
 */
class Negyzet : public Alakzat {
    private:
        double a;
    public:
        Negyzet(unsigned szin, double x, double y, double a)
            : Alakzat(szin, x, y), a(a) {}
        
        virtual void kiir() const {
            std::cout << "Négyzet x=" << x << " y=" << y
                      << " a=" << a << std::endl;
        }
        
        virtual double kerulet() const {
            return 4*a;
        }

        virtual Alakzat *klon() const {
            return new Negyzet(*this);
        }
};

class Rajztabla {
    private:
        std::vector<Alakzat*> alakzatok;
    public:
        Rajztabla() {}
    
        ~Rajztabla() {
            for (size_t i = 0; i < alakzatok.size(); ++i)
                delete alakzatok[i];
        }
        
        Rajztabla(Rajztabla const & r) {
            for (size_t i = 0; i < r.alakzatok.size(); ++i)
                alakzatok.push_back(r.alakzatok[i]->klon());
        }
        
        /* kéne, de lusta voltam */
        Rajztabla& operator=(Rajztabla const & r);
        
        void hozzaad(Alakzat *a) {
            alakzatok.push_back(a);
        }
        
        void torol(size_t i) {
            delete alakzatok[i];
            alakzatok.erase(alakzatok.begin() + i);
        }
        
        void mindet_rajzol() const {
            for (size_t i = 0; i < alakzatok.size(); ++i)
                alakzatok[i]->kiir();
        }
        
        double osszkerulet() const {
            double k = 0;
            for (size_t i = 0; i < alakzatok.size(); ++i)
                k += alakzatok[i]->kerulet();
            return k;
        }
};

/* itt nagyon látszik, hogy kell az alakzat virtuális destruktor:
 * ennek a vektora dinamikus! */
struct Vektor {
    double x, y;
    Vektor() {}
    Vektor(double x, double y) : x(x), y(y) {}
    Vektor operator-(Vektor v) const {
        return Vektor(x-v.x, y-v.y);
    }
    double hossz() const {
        return sqrt(x*x + y*y);
    }
};

class Poligon : public Alakzat {
    private:
        std::vector<Vektor> csucsok;
    public:
        Poligon(unsigned szin, double x, double y, std::vector<double> koord)
            : Alakzat(szin, x, y) {
            if (koord.size() % 2 != 0)
                throw std::invalid_argument("számpárok kellenek");
            if (koord.size()/2 < 3)
                throw std::invalid_argument("sokszöghöz legalább 3 csúcs kell");
            for (size_t i = 0; i < koord.size(); i += 2)
                csucsok.push_back(Vektor(koord[i], koord[i+1]));
        }
        
        virtual void kiir() const {
            std::cout << "Poligon x=" << x << " y=" << y << "..." << std::endl;
        }
        
        virtual double kerulet() const {
            double k = 0;
            for (size_t i = 0; i < csucsok.size()-1; i += 1)
                k += (csucsok[i] - csucsok[i+1]).hossz();
            k += (csucsok[csucsok.size()-1] - csucsok[0]).hossz();
            return k;
        }

        virtual Alakzat *klon() const {
            return new Poligon(*this);
        }
};

int main() {
    Rajztabla r;

    r.hozzaad(new Teglalap(0xFF0000, 100, 20, 50, 60));
    r.hozzaad(new Kor(0x00FFFF, 50, 50, 100));
    r.mindet_rajzol();
    std::cout << r.osszkerulet() << std::endl;
    r.torol(0);
    std::cout << "===" << std::endl;

    r.mindet_rajzol();
    std::cout << r.osszkerulet() << std::endl;
    std::cout << "===" << std::endl;
    
    std::vector<double> csucsok = { 10, 20, 30, 40, 50, 60 };
    r.hozzaad(new Poligon(0xFF00FF, 5, 6, csucsok));
    r.mindet_rajzol();
    std::cout << r.osszkerulet() << std::endl;
    std::cout << "===" << std::endl;

    Rajztabla r2 = r;
    r2.mindet_rajzol();
    std::cout << "===" << std::endl;
}
